@@ -19,20 +19,36 @@
    Boston, MA 02110-1301, USA.
 */
 
-#include <kservicegroup.h>
-#include <qmimedatabase.h>
-#include <QtTest>
+#include <QtTest/QTest>
+#include <QtTest/QSignalSpy>
+
+#include <QMimeDatabase>
+#include <QThread>
+#include <QTimer>
+#include <QStandardPaths>
+#include <QDebug>
+#include <QFile>
+#include <kbuildsycoca_p.h>
 
 #include <kconfig.h>
 #include <kconfiggroup.h>
 #include <kdesktopfile.h>
 #include <ksycoca.h>
-
+#include <kservicegroup.h>
 #include <kservicetype.h>
 #include <kservicetypeprofile.h>
+#include <QMutex>
 
-#include <qstandardpaths.h>
-#include <QDebug>
+
+static QString fakeTextPluginDesktopFile()
+{
+    return QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1String("/kservices5/") + "threadtextplugin.desktop";
+}
+
+static QString fakeServiceDesktopFile()
+{
+    return QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1String("/kservices5/") + "threadfakeservice.desktop";
+}
 
 // Helper method for all the trader tests
 static bool offerListHasService(const KService::List &offers,
@@ -52,6 +68,15 @@ static bool offerListHasService(const KService::List &offers,
     return found;
 }
 
+static QSet<QThread *> s_threadsWhoSawFakeService;
+static QMutex s_setMutex;
+static int threadsWhoSawFakeService() {
+    QMutexLocker locker(&s_setMutex);
+    return s_threadsWhoSawFakeService.count();
+}
+static QAtomicInt s_fakeServiceDeleted = 0;
+
+
 class WorkerObject : public QObject
 {
     Q_OBJECT
@@ -61,7 +86,7 @@ public:
 public Q_SLOTS:
     void work()
     {
-        qDebug() << QThread::currentThread() << "working...";
+        //qDebug() << QThread::currentThread() << "working...";
 
         const KServiceType::List allServiceTypes = KServiceType::allServiceTypes();
         Q_ASSERT(!allServiceTypes.isEmpty());
@@ -84,6 +109,12 @@ public Q_SLOTS:
             Q_ASSERT(!entryPath.isEmpty());
 
             KService::Ptr lookedupService = KService::serviceByDesktopPath(entryPath);
+            if (!lookedupService) {
+                if (entryPath == "threadfakeservice.desktop" && s_fakeServiceDeleted) { // ok, it got deleted meanwhile
+                    continue;
+                }
+                qWarning() << entryPath << "is gone!";
+            }
             Q_ASSERT(lookedupService);   // not null
             QCOMPARE(lookedupService->entryPath(), entryPath);
 
@@ -94,20 +125,31 @@ public Q_SLOTS:
                 }
                 Q_ASSERT(!menuId.isEmpty());
                 lookedupService = KService::serviceByMenuId(menuId);
+                if (!lookedupService) {
+                    if (menuId == "threadfakeservice" && s_fakeServiceDeleted) { // ok, it got deleted meanwhile
+                        continue;
+                    }
+                    qWarning() << menuId << "is gone!";
+                }
                 Q_ASSERT(lookedupService);   // not null
                 QCOMPARE(lookedupService->menuId(), menuId);
             }
         }
 
         KService::List offers = KServiceTypeTrader::self()->query("KPluginInfo");
-        Q_ASSERT(offerListHasService(offers, "faketextplugin.desktop"));
+        Q_ASSERT(offerListHasService(offers, "threadtextplugin.desktop"));
 
-        offers = KServiceTypeTrader::self()->query("KPluginInfo", "Library == 'faketextplugin'");
+        offers = KServiceTypeTrader::self()->query("KPluginInfo", "Library == 'threadtextplugin'");
         Q_ASSERT(offers.count() == 1);
-        QVERIFY(offerListHasService(offers, "faketextplugin.desktop"));
+        QVERIFY(offerListHasService(offers, "threadtextplugin.desktop"));
 
         KServiceGroup::Ptr root = KServiceGroup::root();
         Q_ASSERT(root);
+
+        if (KService::serviceByDesktopPath("threadfakeservice.desktop")) {
+            QMutexLocker locker(&s_setMutex);
+            s_threadsWhoSawFakeService.insert(QThread::currentThread());
+        }
     }
 };
 
@@ -131,7 +173,7 @@ public:
         m_stop = true;
     }
 private:
-    bool m_stop;
+    QAtomicInt m_stop; // bool
 };
 
 /**
@@ -159,21 +201,16 @@ public:
     }
 };
 
-extern KSERVICE_EXPORT bool kservice_require_kded;
 
+// This code runs in the main thread
 class KSycocaThreadTest : public QObject
 {
     Q_OBJECT
-public:
-    KSycocaThreadTest()
-    {
-        kservice_require_kded = false;
-    }
-    void launch();
 
 private Q_SLOTS:
     // Note that this isn't a QTest based test.
     // All these methods are called manually.
+    void initTestCase();
     void cleanupTestCase();
     void testCreateService();
     void testDeleteService()
@@ -202,32 +239,26 @@ private:
 
 static void runKBuildSycoca()
 {
-    QProcess proc;
-    const QString kbuildsycoca = QStandardPaths::findExecutable(KBUILDSYCOCA_EXENAME);
-    QVERIFY(!kbuildsycoca.isEmpty());
-    QStringList args;
-    args << "--testmode";
-    proc.setProcessChannelMode(QProcess::MergedChannels); // silence kbuildsycoca output
-    proc.start(kbuildsycoca, args);
-    qDebug() << "waiting for signal";
     QSignalSpy spy(KSycoca::self(), SIGNAL(databaseChanged(QStringList)));
-    QVERIFY(spy.wait(10000));
-
+    KBuildSycoca builder;
+    QVERIFY(builder.recreate());
+    qDebug() << "waiting for signal";
+    QVERIFY(spy.wait(20000));
     qDebug() << "got signal";
-    proc.waitForFinished();
-    QCOMPARE(proc.exitStatus(), QProcess::NormalExit);
 }
 
-void KSycocaThreadTest::launch()
+void KSycocaThreadTest::initTestCase()
 {
+    QStandardPaths::enableTestMode(true);
+
     // This service is always there. Used in the trader queries from the thread.
-    const QString fakeTextPlugin = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1String("/kservices5/") + "faketextplugin.desktop";
+    const QString fakeTextPlugin = fakeTextPluginDesktopFile();
     if (!QFile::exists(fakeTextPlugin)) {
         KDesktopFile file(fakeTextPlugin);
         KConfigGroup group = file.desktopGroup();
-        group.writeEntry("Name", "FakeTextPlugin");
+        group.writeEntry("Name", "ThreadTextPlugin");
         group.writeEntry("Type", "Service");
-        group.writeEntry("X-KDE-Library", "faketextplugin");
+        group.writeEntry("X-KDE-Library", "threadtextplugin");
         group.writeEntry("X-KDE-Protocols", "http,ftp");
         group.writeEntry("ServiceTypes", "KPluginInfo");
         group.writeEntry("MimeType", "text/plain;");
@@ -236,20 +267,20 @@ void KSycocaThreadTest::launch()
         runKBuildSycoca();
         // Process the event
         int count = 0;
-        while (!KService::serviceByDesktopPath("faketextplugin.desktop")) {
+        while (!KService::serviceByDesktopPath("threadtextplugin.desktop")) {
             qApp->processEvents();
             if (++count == 20) {
-                qFatal("sycoca doesn't have faketextplugin.desktop");
+                qFatal("sycoca doesn't have threadtextplugin.desktop");
             }
         }
     }
 
     // Start clean
-    const QString servPath = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1String("/kservices5/") + "fakeservice.desktop";
+    const QString servPath = fakeServiceDesktopFile();
     if (QFile::exists(servPath)) {
         QFile::remove(servPath);
     }
-    if (KService::serviceByDesktopPath("fakeservice.desktop")) {
+    if (KService::serviceByDesktopPath("threadfakeservice.desktop")) {
         deleteFakeService();
     }
     threads.resize(5);
@@ -257,65 +288,85 @@ void KSycocaThreadTest::launch()
         threads[i] = i < 3 ? new WorkerThread : new EventLoopThread;
         threads[i]->start();
     }
-    testCreateService();
 }
 
 void KSycocaThreadTest::cleanupTestCase()
 {
-    const QString fakeTextPlugin = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1String("/kservices5/") + "faketextplugin.desktop";
-    QFile::remove(fakeTextPlugin);
+    QFile::remove(fakeTextPluginDesktopFile());
+}
+
+// duplicated from kcoreaddons/autotests/kdirwatch_unittest.cpp
+static void waitUntilAfter(const QDateTime &ctime)
+{
+    int totalWait = 0;
+    QDateTime now;
+    Q_FOREVER {
+        now = QDateTime::currentDateTime();
+        if (now.toTime_t() == ctime.toTime_t())   // truncate milliseconds
+        {
+            totalWait += 50;
+            QTest::qWait(50);
+        } else {
+            QVERIFY(now > ctime); // can't go back in time ;)
+            QTest::qWait(50); // be safe
+            break;
+        }
+    }
+    //if (totalWait > 0)
+    qDebug() << "Waited" << totalWait << "ms so that now" << now.toString(Qt::ISODate) << "is >" << ctime.toString(Qt::ISODate);
 }
 
 void KSycocaThreadTest::testCreateService()
 {
-    createFakeService();
-    const QString servPath = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1String("/kservices5/") + "fakeservice.desktop";
+    // Wait one second so that ksycoca can detect a mtime change
+    // ## IMHO this is a Qt bug, QFileInfo::lastModified() should include milliseconds
+    waitUntilAfter(QDateTime::currentDateTime());
 
-    Q_ASSERT(QFile::exists(servPath));
+    createFakeService();
+    QVERIFY(QFile::exists(fakeServiceDesktopFile()));
     qDebug() << "executing kbuildsycoca (1)";
     runKBuildSycoca();
 
-    QTimer::singleShot(1000, this, SLOT(testDeleteService()));
+    QTRY_VERIFY(KService::serviceByDesktopPath("threadfakeservice.desktop"));
+
+    // Now wait to check that all threads saw that new service
+    QTRY_COMPARE_WITH_TIMEOUT(threadsWhoSawFakeService(), threads.size(), 20000);
 }
 
 void KSycocaThreadTest::deleteFakeService()
 {
-    KService::Ptr fakeService = KService::serviceByDesktopPath("fakeservice.desktop");
-    Q_ASSERT(fakeService);
-    const QString servPath = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1String("/kservices5/") + "fakeservice.desktop";
+    s_fakeServiceDeleted = 1;
+
+    qDebug() << "now deleting the fake service";
+    KService::Ptr fakeService = KService::serviceByDesktopPath("threadfakeservice.desktop");
+    QVERIFY(fakeService);
+    const QString servPath = fakeServiceDesktopFile();
     QFile::remove(servPath);
 
     QSignalSpy spy(KSycoca::self(), SIGNAL(databaseChanged(QStringList)));
-    Q_ASSERT(spy.isValid());
+    QVERIFY(spy.isValid());
 
     qDebug() << "executing kbuildsycoca (2)";
     runKBuildSycoca();
-    Q_ASSERT(spy[0][0].toStringList().contains("services"));
+    QVERIFY(!spy.isEmpty());
+    QVERIFY(spy[0][0].toStringList().contains("services"));
 
-    Q_ASSERT(fakeService); // the whole point of refcounting is that this KService instance is still valid.
-    Q_ASSERT(!QFile::exists(servPath));
+    QVERIFY(fakeService); // the whole point of refcounting is that this KService instance is still valid.
+    QVERIFY(!QFile::exists(servPath));
 }
 
 void KSycocaThreadTest::createFakeService()
 {
-    const QString fakeService = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1String("/kservices5/") + "fakeservice.desktop";
-    KDesktopFile file(fakeService);
+    KDesktopFile file(fakeServiceDesktopFile());
     KConfigGroup group = file.desktopGroup();
-    group.writeEntry("Name", "FakeService");
+    group.writeEntry("Name", "ThreadFakeService");
     group.writeEntry("Type", "Service");
-    group.writeEntry("X-KDE-Library", "fakeservice");
+    group.writeEntry("X-KDE-Library", "threadfakeservice");
     group.writeEntry("X-KDE-Protocols", "http,ftp");
     group.writeEntry("ServiceTypes", "KPluginInfo");
     group.writeEntry("MimeType", "text/plain;");
 }
 
-int main(int argc, char **argv)
-{
-    QStandardPaths::enableTestMode(true);
-    QCoreApplication app(argc, argv);
-    KSycocaThreadTest mainObj;
-    mainObj.launch();
-    return app.exec();
-}
 
+QTEST_MAIN(KSycocaThreadTest)
 #include "ksycocathreadtest.moc"
